@@ -1,7 +1,7 @@
 // Copyright 2022 David Robillard <d@drobilla.net>
 // SPDX-License-Identifier: ISC
 
-#include "sratom/sratom.h" // IWYU pragma: keep
+#include <sratom/sratom.h> // IWYU pragma: keep
 
 #ifdef __GNUC__
 __attribute__((const))
