@@ -9,6 +9,7 @@
 #include <serd/serd.h>
 #include <sratom/sratom.h>
 
+#include <assert.h>
 #include <stdarg.h>
 #include <stdbool.h>
 #include <stdint.h>
@@ -37,6 +38,7 @@ copy_string(const char* str)
 {
   const size_t len = strlen(str);
   char*        dup = (char*)malloc(len + 1);
+  assert(dup);
   memcpy(dup, str, len + 1);
   return dup;
 }
@@ -81,7 +83,7 @@ test_fail(const char* fmt, ...)
 }
 
 static int
-test(SerdEnv* env, bool top_level, bool pretty_numbers)
+test(SerdEnv* env, const char* base_uri, bool top_level, bool pretty_numbers)
 {
   Uris           uris  = {NULL, 0};
   LV2_URID_Map   map   = {&uris, urid_map};
@@ -319,8 +321,6 @@ test(SerdEnv* env, bool top_level, bool pretty_numbers)
 
   lv2_atom_forge_pop(&forge, &obj_frame);
 
-  const char* base_uri = "file:///tmp/base/";
-
   SerdNode s = serd_node_from_string(SERD_URI, USTR("http://example.org/obj"));
   SerdNode p = serd_node_from_string(SERD_URI, USTR(NS_RDF "value"));
 
@@ -384,10 +384,11 @@ test(SerdEnv* env, bool top_level, bool pretty_numbers)
 static int
 test_env(SerdEnv* env)
 {
-  if (test(env, false, false) || //
-      test(env, true, false) ||  //
-      test(env, false, true) ||  //
-      test(env, true, true)) {
+  if (test(env, "file:///tmp/base/", false, false) || //
+      test(env, "file:///tmp/base/", true, false) ||  //
+      test(env, "file:///tmp/base/", false, true) ||  //
+      test(env, "file:///tmp/base/", true, true) ||   //
+      test(env, "http://example.org/", true, true)) {
     return 1;
   }
 
